import numpy as np
from . import calculation


def projection(x, index, index1=None):
    """
    projection of state or operator in subspace {index}
    :param x: state or operator
    :param index: bases of subspace
    :param index1: bases of another subspace
    :return: projected state or operator
    """
    if len(x.shape) < 2:
        n = len(x)
        proj = np.eye(n)[index]
        return np.dot(proj, x)
    else:
        n = len(x[0])
        proj = np.eye(n)[index]
        if index1:
            proj1 = np.eye(n)[index1]
            return np.dot(np.dot(proj, x), proj1.T)
        else:
            return np.dot(np.dot(proj, x), proj.T)


def normalize(state):
    """
    normalization of state
    :param state:
    :return: np.array
    """
    tmp = state/np.linalg.norm(state)
    return tmp


def dagger(x):
    """
    Hermitian conjugate state
    :param x: state
    :return: Hermitian conjugate state
    """
    return x.T.conj()


def ketbra(state1, state2):
    """
    out product of state1 and state2
    :type state1: np.array
    :type state2: np.array
    :return: complex number
    """
    state1 = normalize(state1)
    state2 = normalize(state2)
    return np.outer(state1.conj(), state2)


def braket(state1, state2):
    """
    inner product of state1 and state2
    :type state1: np.array
    :type state2: np.array
    :return: np.array
    """
    state1 = normalize(state1)
    state2 = normalize(state2)
    return np.dot(state1.conj(), state2)


def density(ensembles):
    """
    density matrix of a ensemble of quantum states
    :param ensembles: np.array
    :return: np.array
    """
    if len(ensembles.shape) < 2:
        return ketbra(ensembles)
    else:
        den_mat = ketbra(ensembles[0])
        for i in range(1, len(ensembles)):
            den_mat += ketbra(ensembles[i])
        den_mat /= len(ensembles)
        return den_mat


def fidelity(state1, state2):
    """
    fidelity between state1 and state2, only valid for pure state1
    :type state1: np.array
    :type state2: np.array
    :return: real number
    """
    if len(state1.shape) > 1:
        print("error: state1 must be a pure state.")
    state1 = normalize(state1)
    fid = 0
    if len(state2.shape)<2:
        state2 = normalize(state2)
        fid = np.abs(
            braket(state1, state2)
        )**2
    else:
        for i in range(len(state2)):
            state2[i] = normalize(state2[i])
            fid += np.abs(
                braket(state1, state2)
            )**2
        fid /= len(state2)
    return fid.real


def purity(den_mat):
    """
    purity of density matrix
    :type den_mat: np.array
    :return: real number
    """
    return (np.trace(
        np.dot(den_mat, den_mat)
    )).real


def trans_base(bases, state, new_way=False):
    """
    Base transition.
    :param new_way: whether use the fast way or not
    :param bases: bases of each site
    :type bases: string
    :param state: quantum state
    :return: transformed quantum state
    """
    Z2Z = np.eye(2)
    Z2X = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
    Z2Y = 1 / np.sqrt(2) * np.array([[1, -1j], [1, 1j]])
    decode = {'Z': Z2Z, 'X': Z2X, 'Y': Z2Y, 'z': Z2Z, 'x': Z2X,'y': Z2Y}
    if not new_way:
        tmp_mat = decode[bases[0]]
        for i in range(1, len(bases)):
            tmp_mat = np.kron(tmp_mat, decode[bases[i]])
        return np.dot(tmp_mat, state)
    else:
        for i in range(len(bases)):
            if bases[i] not in ['z', 'Z']:
                state = calculation.fast_dot(i, decode[bases[i]], state)
        return state


# def subsystem(sites, x):
#     """
#     density matrix of subsystem consisted of sites
#     :param sites: region of interest
#     :param x: state or density matrix
#     :return: reduced density matrix
#     """















import numpy as np
from . import utils, gate, measurement
from .calculation import fast_dot


class Circuit:
    def __init__(self, N, sites, gates, paras, ideal=True, noise_fuc=None):
        """
        :param N: length of atoms chain
        :type sites: list of list(turtle)
        :type gates: list of string
        :type paras: dictionary from string to list(turtle)
        :type ideal: True or False
        :param noise_fuc: the function to generate noised parameters
        """
        self._N = N
        self.sites = sites
        self.gates = gates
        self.ideal = ideal
        self.paras = paras
        self.noise_fuc = noise_fuc
        self.steps = len(sites)
        self.init_state = dict()
        self.layer = []
        self.forward = []
        self.backward = []
        self.final_state = self.init_state

    @property
    def N(self):
        return self._N

    @property
    def space_size(self):
        """
        size of Hilbert space
        :return: int
        """
        return 2 ** self.N

    def hilbert_space(self):
        """
        Hilbert space
        :return: np.array
        """
        return np.arange(self.space_size)

    def state_init(self, state):
        """
        initialize state
        :param state: bit string of the site or a state vector
        :type state: str/np.array
        :return: np.array
        """
        if isinstance(state, str):
            self.init_state[utils.loc(state)] = 1+0j
            self.final_state = self.dic2array(self.init_state)
            self.trees(state)
        else:
            self.init_state = state
            self.final_state = self.init_state

    def dic2array(self, x):
        """
        a tool to transform a quantum state stored in a dictionary to a np.array
        :param x: quantum state
        :type x: dictionary
        :return: the transformed quantum state
        :rtype: np.array
        """
        state = np.zeros(self.space_size, dtype='complex')
        for i in x:
            state[i] = x[i]
        return state

    def trees(self, state):
        """
        build forward calculation tree\backward calculation tree and the layers
        :param state: initial state
        :type state: str
        :return: None
        """
        layer_tmp = [utils.loc(state)]
        self.layer.append(layer_tmp)

        gates_shape = {}
        for op in self.paras:
            opc = gate.gate_map[op]
            paras = self.paras[op]
            gates_shape[op] = opc(self.ideal, *paras).gate_shape

        for step in range(self.steps):
            sites = self.sites[step]
            n = len(sites)

            mat_shape = gates_shape[self.gates[step]]
            layer_tmp = set()
            map_tmp_f = dict()
            map_tmp_b = dict()

            for node in self.layer[-1]:
                node_index = utils.index(node, self.N, sites)
                sons = set()
                for base in range(2 ** n):
                    if mat_shape[base, node_index]:
                        tmp = node
                        for site in range(n):
                            tmp = utils.set_bit(
                                tmp, sites[site], self.N, utils.site_bit(
                                    base, n, site+1)
                            )
                        sons.add(tmp)
                map_tmp_f[node] = sons

                layer_tmp = layer_tmp.union(sons)
                for son in sons:
                    if son in map_tmp_b:
                        map_tmp_b[son].add(node)
                    else:
                        parent = set()
                        parent.add(node)
                        map_tmp_b[son] = parent

            self.layer.append(list(layer_tmp))
            self.forward.append(map_tmp_f)
            self.backward.insert(0, map_tmp_b)

    def branch(self, x):
        """
        generate a calculation branch of element x
        :param x: the x-th base of quantum state
        :return: the branch
        :rtype: a list of list
        """
        node_tmp = set()
        node_tmp.add(x)
        bran = [list(node_tmp)]
        for i in range(self.steps):
            node_tmp = set()
            for j in bran[0]:
                node_tmp = node_tmp.union(self.backward[i][j])
            bran.insert(0, list(node_tmp))
        return bran

    def state_generation(self):
        """
        generate a quantum state
        :return: generated quantum state
        """
        gates = dict()
        for op in self.paras:
            opc = gate.gate_map[op]
            paras = self.paras[op]
            gates[op] = opc(self.ideal, *paras)
        tmp_state = self.init_state
        for step in range(self.steps):
            sites = self.sites[step]
            op = gates[self.gates[step]].generator()
            if isinstance(tmp_state, dict):
                tmp_state = fast_dot(sites, op, tmp_state, self.N, self.layer[step+1])
                self.final_state = self.dic2array(tmp_state)
            else:
                tmp_state = fast_dot(sites, op, tmp_state, self.N)
                self.final_state = tmp_state
        return tmp_state

    def element(self, x):
        """
        calculate the x-th element of the final state
        :param x: base
        :return: psi(x), psi is the final state
        """
        branch = self.branch(x)
        gates = dict()
        for op in self.paras:
            opc = gate.gate_map[op]
            paras = self.paras[op]
            gates[op] = opc(self.ideal, *paras)
        tmp_state = self.init_state
        for step in range(self.steps):
            sites = self.sites[step]
            op = gates[self.gates[step]].generator()
            tmp_state = fast_dot(sites, op, tmp_state, self.N, branch[step+1])
        return tmp_state[branch[-1][0]]

    def ensemble_generation(self, size):
        """
        generate a ensemble of quantum state
        :param size: size of ensemble
        :return: ensembles of quantum state
        """
        ensemble = []
        for _ in range(size):
            ensemble.append(self.state_generation())
        ensemble = np.array(ensemble)
        self.final_state = ensemble
        return ensemble

    def state_amp(self, state=None):
        """
        the positive and real state of target state
        :param state: target state
        :return: the positive and real state of target state
        """
        state = state if isinstance(state, np.ndarray) else self.final_state
        amp_state = np.array(
            [np.abs(state[i]) for i in range(len(state))],
            dtype='complex')
        return amp_state

    def show(self, state=None):
        state = state if isinstance(state, np.ndarray) else self.final_state
        for i in range(len(state)):
            if np.abs(state[i]) > 1e-10:
                print('|'+utils.int2bin(i, self.N)+'>',
                      round(np.abs(state[i]), 3),
                      str(round(np.angle(state[i], False)/np.pi, 3))+'pi',
                      round(state[i].real, 3)+1j*round(state[i].imag, 3)
                      )

    def state_save(self, state=None, path=None):
        """
        save quantum state at path
        :param state:
        :param path:
        :return: None
        """
        state = state if isinstance(state, np.ndarray) else self.final_state
        path = path if path else str(self.N) + ' qubits_state' + '.txt'
        f = open(path, mode='w')
        for i in range(self.space_size):
            f.write(str(state[i].real) + ' ' + str(state[i].imag) + '\n')
        f.close()
        print('state saved at ' + path)

    def amp_save(self, state, path=None):
        """
        sate the positive and real state of state at path
        :param state:
        :param path:
        :return: None
        """
        state = state if isinstance(state, np.ndarray) else self.final_state
        path = path if path else str(self.N) + ' qubits_state_amp' + '.txt'
        amp_state = self.state_amp(state)
        self.state_save(amp_state, path)

    def mc_sample(self, size, error=0):
        """
        Monte Carlo method to generate samples
        :param size: size of samples
        :param error: measurement error
        :return: measurement results, (measurement results with measurement error)
        """
        self.state_generation()
        cover_func = 2 * measurement.distribution(self.final_state)
        x_list = self.sample(n=int(2.5*size), state=self.final_state)
        y_list = np.random.uniform(size=int(2.5*size))          
        count = 0
        measure = []
        for i in range(len(x_list)):
            y0 = self.element(x_list[i])
            if cover_func[x_list[i]] * y_list[i] < np.abs(y0)**2:
                count += 1
                measure.append(x_list[i])
                if count == size:
                    break
        if not error:
            return np.array(measure)
        else:
            return np.array(measure), measurement.sample_error(
                np.array(measure), self.N, error
            )

    def sample(self, n=1, state=None, bases=None, error=0):
        """
        generate samples
        :param n: size of smaples
        :param state: quantum state
        :param bases: measurement bases
        :param error: measurement error
        :return: sample results
        """
        state = state if isinstance(state, np.ndarray) else self.final_state
        if len(state.shape) < 2:
            return measurement.sample(state, n, bases, error)
        else:
            return measurement.ensemble_sample(state, error)



import numpy as np
from . import utils


# 反向计算 快
def fast_dot(sites, gate, state, N, layer=None):
    """
    A fast way to calculate the unitary transformation acting on some sites.

    :param sites: the sites on where the unitary transformation act
    :type sites: array like
    :param gate: the local transformation
    :param state: init quantum state
    :type state: dictionary
    :param layer: a list of index of final state
    :type layer: list
    :return: if state is a dictionary, then return a dictionary, else return a np.array
    """
    n = len(sites)
    layer = layer if layer else range(2 ** N)
    index = [utils.index(i, N, sites) for i in layer]
    gate = gate.T
    trans_mat = gate[:, index]
    v = np.array(layer).reshape(1, len(layer)).repeat(2 ** n, 0)
    for base in range(2 ** n):
        for site in range(n):
            v[base, :] = utils.set_bit(
                v[base, :], sites[site], N, utils.site_bit(base, n, site + 1)
            )

    if isinstance(state, dict):
        v = v.astype('complex')
        for i in range(len(v)):
            for j in range(len(v[i])):
                v[i, j] = state.get(v[i, j], 0)

        trans_mat *= v
        res = trans_mat.sum(0)
        final_state = dict(zip(layer, res))
        return final_state
    else:
        v = state[v]
        trans_mat *= v
        res = trans_mat.sum(0)
        return res


# 前向计算0 慢
def fast_dot_(sites, gate, state, N, layer=None):
    """
    A fast way to calculate the unitary transformation acting on some sites.

    :param sites: the sites on where the unitary transformation act
    :type sites: array like
    :param gate: the local transformation
    :param state: init quantum state
    :type state: dictionary
    :param layer: a list of index of final state
    :type layer: list
    :return: if state is a dictionary, then return a dictionary, else return a np.array
    """
    n = len(sites)
    layer = layer if layer else range(2 ** N)
    index = [utils.index(i, N, sites) for i in layer]
    trans_mat = gate[:, index]
    v = np.array(layer).reshape(1, len(layer)).repeat(2 ** n, 0)
    for base in range(2 ** n):
        for site in range(n):
            v[base, :] = utils.set_bit(
                v[base, :], sites[site], N, utils.site_bit(base, n, site + 1)
            )

    if isinstance(state, dict):
        v0 = np.array([state.get(x, 0) for x in layer])
        trans_mat *= v0
        final_state = dict()
        for i in range(len(trans_mat)):
            for j in range(len(trans_mat[i])):
                final_state[v[i, j]] = final_state.setdefault(v[i, j], 0) + trans_mat[i, j]
        return final_state
    else:
        v = state[v]
        trans_mat *= v
        res = trans_mat.sum(0)
        return res


# 前向计算 快
def fast_dot_2(sites, gate, state, N, layer=None, relation=None):
    """
    A fast way to calculate the unitary transformation acting on some sites.

    :param sites: the sites on where the unitary transformation act
    :type sites: array like
    :param gate: the local transformation
    :param state: init quantum state
    :type state: dictionary
    :param layer: a list of index of final state
    :type layer: list
    :return: if state is a dictionary, then return a dictionary, else return a np.array
    """
    # n = len(sites)
    layer = layer if layer else range(2 ** N)
    final_state = dict()
    for node in layer:
        index = utils.index(node, N, sites)
        for v in relation[node]:
            index1 = utils.index(v, N, sites)
            final_state[v] = final_state.setdefault(v, 0) + state[node] * gate[index1, index]
    return final_state


def fast_dot_1(sites, gate, state, N, layer=None, relation=None):
    """
    A fast way to calculate the unitary transformation acting on some sites.

    :param sites: the sites on where the unitary transformation act
    :type sites: array like
    :param gate: the local transformation
    :param state: init quantum state
    :type state: dictionary
    :param layer: a list of index of final state
    :type layer: list
    :return: if state is a dictionary, then return a dictionary, else return a np.array
    """
    # n = len(sites)
    layer = layer if layer else range(2 ** N)
    final_state = dict()
    for node in layer:
        index = utils.index(node, N, sites)
        for v in relation[node]:
            index1 = utils.index(v, N, sites)
            final_state[v] = final_state.setdefault(v, 0) + state[node] * gate[index1, index]
    return final_state





import numpy as np
import scipy.stats as stats
from scipy.stats import bernoulli
from . import unitary, utils
# from .circuit import Circuit
import random


def distribution(x):
    """
    non-selective Z-measurement distribution
    :param x: state
    :type x: np.array
    :return: np.array
    """
    x /= np.linalg.norm(x, 2)
    prob = np.zeros(x.size)
    for i in range(len(x)):
        prob[i] = np.abs(x[i]) ** 2
    return prob


def dis2state(x):
    """
    turn a distribution into positive and real quantum state
    :param x: distribution
    :type x: np.array
    :return: np.array
    """
    state = np.array([np.sqrt(x[i]) for i in range(len(x))])
    return state


# TODO 添加多bases测量
def sample(state, n=1, bases=None, error=0):
    """
    n times measurement for quantum state.
    :param bases: measurement bases
    :param state: quantum state
    :param n: size of sample
    :return: n times measurement and its distribution
    :param error: measurement errors
    """
    N = utils.length(state)
    if not bases:
        measure = stats.rv_discrete(
            values=(range(len(state)), distribution(state))
        ).rvs(size=n)
        if not error:
            return measure[:n]
        else:
            return measure[:n], sample_error(measure[:n], N, error)
    elif isinstance(bases, str):
        new_state = unitary.trans_base(bases, state)
        measure = stats.rv_discrete(
            values=(range(len(new_state)), distribution(new_state))
        ).rvs(size=n)
        if not error:
            return measure[:n]
        else:
            return measure[:n], sample_error(measure[:n], N, error)
    else:
        measure = []
        if error:
            measure_error = []
        for base in bases:
            new_state = unitary.trans_base(base, state)
            measure.append(
                stats.rv_discrete(
                    values=(range(len(new_state)), distribution(new_state))
                ).rvs(size=n)
            )
            if error:
                measure_error.append(sample_error(measure[-1]), N, error)
        if not error:
            return np.array(measure)
        else:
            return np.array(measure), np.array(measure_error)


def ensemble_sample(ensembles, error=0):
    """
    one measurement for each state of an ensemble
    :param error: measurement
    :param ensembles: measurement ensembles
    :return: measurement results
    """
    N = utils.length(ensembles[0])
    measurement = []
    for tmp in ensembles:
        measurement.append(sample(tmp))
    if not error:
        return np.array(measurement)
    else:
        return np.array(measurement), sample_error(measurement, N, error)


def sample_error(samples, N, error=0):
    """
    Simulate the measurement error, each bit has a probability of error to be
    flipped. (first order approximation be used)
    :param samples: measurement results
    :param N: length of atoms chain
    :param error: probability of flips
    :return: measurement results with errors
    """
    size = len(samples)
    flip = bernoulli.rvs(N * error, size=size)
    ker = [2 ** i for i in range(N)]
    count = 0
    for i in range(size):
        if flip[i]:
            count += 1
            flip[i] = random.choice(ker)
    print("errors({0}), error rate = {1}.".format(count, count / size))
    sample_new = np.array([samples[i] ^ flip[i] for i in range(size)])
    return sample_new


def sample_save(samples, N, path=None):
    """
    save samples at path in 0/1 series
    :param samples: measurement results
    :param N: length of atoms chain
    :param path: path to be saved
    :return: None
    """
    path = path if path else str(N) + ' qubits_measurement.txt'
    f = open(path, mode='w')
    for i in samples:
        tmp = utils.int2bin(i, N)
        for j in tmp:
            f.write(j + ' ')
        f.write('\n')
    f.close()
    print("saved at " + path)


def bases_save(bases, M, path1=None, path2=None):
    """
    save measurement bases of each measurement at path1
    save the set of measurement bases at path2
    :param bases: the measurement bases
    :param M: measurement times
    :param path1: the path to save each measurement base
    :param path2: the path to save measurement bases set
    :return: None
    """
    N = len(bases[0])
    path1 = path1 if path1 else str(N) + ' qubits_measurement_bases.txt'
    path2 = path2 if path2 else str(N) + ' qubits_measurement_bases_set.txt'
    f1 = open(path1, mode='w')
    f2 = open(path2, mode='w')
    for base in bases:
        for i in base:
            f2.write(i + ' ')
            f2.write('\n')
        for _ in range(M):
            for i in base:
                f1.write(i + ' ')
                f1.write('\n')
    f1.close()
    f2.close()
    print("bases saved at " + path1)
    print("bases set saved at " + path2)


def sample_dis(samples, N):
    """
    turn a bunch of samples into distribution
    :param samples: samples
    :param N: size of total qubits
    :return: np.array
    """
    n = len(samples)
    sam_dis = np.zeros(2 ** N)
    for i in samples:
        sam_dis[i] += 1
    sam_dis /= n
    return sam_dis


def density_dis(x):
    """
    extract the distribution from a density matrix
    :param x:
    :return:
    """
    n = len(x)
    dis = np.array([x[i][i] for i in range(n)])
    return dis.real


def kld(p, q):
    """
    K-L divergence between distribution p and q.
    If some elements of p or q equal to zero, just remove these terms.
    :param p: distribution
    :param q: distribution
    :return: real number
    """
    divergence = 0
    for i in range(len(p)):
        if p[i] and q[i]:
            divergence += p[i] * np.log(p[i] / q[i])
    return divergence


def kld_(p, q):
    """
    K-L divergence between distribution p and q.
    A little trick to handle the zero case. Turn the infinity into a great number.
    :param p: distribution
    :param q: distribution
    :return: real number
    """
    n = len(p)
    epsilon = 0.01 / n
    p = p + epsilon
    q = q + epsilon
    divergence = np.sum(p * np.log(p / q))
    return divergence



import abc

import numpy as np
from scipy.linalg import expm
from . import utils, unitary


def noise(x):
    mu, sigma = x, x / 25.76
    np.random.seed()
    return np.random.normal(mu, sigma, 1)[0]


def sparse_check(x):
    tmp = x.flatten(order='C')
    nonzero = 0
    for i in range(len(tmp)):
        if tmp[i] != 0:
            nonzero += 1
    return nonzero, nonzero / len(tmp)


def unitary_check(x, threshold=1e-7):
    distance = np.linalg.norm(np.dot(unitary.dagger(x), x) - np.eye(len(x)))
    if distance < threshold:
        return True
    else:
        print('not unitary, error = {}'.format(distance))


gate_map = dict()


class Gate(abc.ABC):
    def __init__(self, ideal, *paras):
        self.ideal = ideal
        self.paras = paras
        self.ideal_gate = self.generator_(*self.paras)
        self.gate_shape = (self.ideal_gate != 0).astype('int')

    @abc.abstractmethod
    def generator_(self, *paras):
        pass

    def generator(self):
        if self.ideal:
            return self.ideal_gate
        else:
            paras = [noise(x) for x in self.paras]
            return self.generator_(*paras)


class Swap(Gate):
    def generator_(self, U, J, t, Delta=0):
        H = np.array([[U, -np.sqrt(2) * J, 0, 0, 0, 0, 0, 0, 0, 0],
                      [-np.sqrt(2) * J, Delta, -np.sqrt(2) * J, 0, 0, 0, 0, 0, 0, 0],
                      [0, -np.sqrt(2) * J, U + 2 * Delta, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, Delta, 0, -J, -J, 0, 0, 0],
                      [0, 0, 0, 0, Delta, -J, -J, 0, 0, 0],
                      [0, 0, 0, -J, -J, U, 0, 0, 0, 0],
                      [0, 0, 0, -J, -J, 0, U + 2 * Delta, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, U, -np.sqrt(2) * J, 0],
                      [0, 0, 0, 0, 0, 0, 0, -np.sqrt(2) * J, Delta, -np.sqrt(2) * J],
                      [0, 0, 0, 0, 0, 0, 0, 0, -np.sqrt(2) * J, U + 2 * Delta]])
        evolution = expm(H * 2 * np.pi * t * -1j)
        index = [1, 3, 4, 8]
        swap_mat = unitary.projection(evolution, index)
        swap_mat /= 1j
        return swap_mat


# TODO 将sto改成更general的sto gate
class Sto(Gate):
    def generator_(self, t, Delta):
        """
        only for 01/10 base
        :param t:
        :param Delta:
        :return:
        """
        phase = np.array(
            [[0, 0, 0, 0],
             [0, 0, 0, 0],
             [0, 0, 2 * Delta, 0],
             [0, 0, 0, 0]]
        )
        evolution = expm(phase * 2 * np.pi * t * -1j)
        return evolution


class Dephase(Gate):
    def generator_(self, phase):
        phase_mat = np.diag([1, np.exp(-1j * phase)])
        return phase_mat

    def generator(self):
        amp = self.paras[0]
        phase = (noise(amp) - amp) * 2 * np.pi
        return self.generator_(phase)


gate_map['swap'] = Swap
gate_map['sto'] = Sto
gate_map['dephase'] = Dephase


############################################################################################################
############################################################################################################
############################################################################################################
# def swap(U, J, t, Delta=0):
#     H = np.array([[U, -np.sqrt(2) * J, 0, 0, 0, 0, 0, 0, 0, 0],
#                   [-np.sqrt(2) * J, Delta, -np.sqrt(2) * J, 0, 0, 0, 0, 0, 0, 0],
#                   [0, -np.sqrt(2) * J, U + 2 * Delta, 0, 0, 0, 0, 0, 0, 0],
#                   [0, 0, 0, Delta, 0, -J, -J, 0, 0, 0],
#                   [0, 0, 0, 0, Delta, -J, -J, 0, 0, 0],
#                   [0, 0, 0, -J, -J, U, 0, 0, 0, 0],
#                   [0, 0, 0, -J, -J, 0, U + 2 * Delta, 0, 0, 0],
#                   [0, 0, 0, 0, 0, 0, 0, U, -np.sqrt(2) * J, 0],
#                   [0, 0, 0, 0, 0, 0, 0, -np.sqrt(2) * J, Delta, -np.sqrt(2) * J],
#                   [0, 0, 0, 0, 0, 0, 0, 0, -np.sqrt(2) * J, U + 2 * Delta]])
#     evolution = expm(H * 2 * np.pi * t * -1j)
#     index = [1, 2, 4, 8]
#     swap_mat = unitary.projection(evolution, index)
#     swap_mat /= 1j
#     return swap_mat
#
#
# def sto(t, Delta):
#     """
#     only for 01/10 base
#     :param t:
#     :param Delta:45
#     :return:
#     """
#     phase = np.array(
#         [[0, 0, 0, 0],
#          [0, 0, 0, 0],
#          [0, 0, 2 * Delta, 0],
#          [0, 0, 0, 0]]
#     )
#     evolution = expm(phase * 2 * np.pi * t * -1j)
#     return evolution
#
#
# swap_shape = np.array(
#     [[1, 0, 0, 0],
#      [0, 1, 1, 0],
#      [0, 1, 1, 0],
#      [0, 0, 0, 1]]
# )
#
# sto_shape = np.diag([1 for _ in range(4)])
#
# gate_shape = dict()
# gate_shape['swap'] = swap_shape
# gate_shape['sto'] = sto_shape
#
#
# def swap_noise(U, J, t, Delta=0):
#     return swap(noise(U), noise(J), noise(t), noise(Delta))
#
#
# def sto_noise(t, Delta):
#     return sto(noise(t), noise(Delta))
#
#
# def NumberOf1(n):
#     count = 0
#     while n & 0xffffffff != 0:
#         count += 1
#         n = n & (n - 1)
#     return count
#
#
# def phase_shift(n):
#     phase = 0
#     for _ in range(n):
#         phase += noise(2 * np.pi)
#     return phase
#
#
# def dephase(n):
#     numbers = []
#     for i in range(2 ** n):
#         numbers.append(NumberOf1(i))
#     numbers = np.array(numbers)
#     dephase_mat = np.diag([np.exp(phase_shift(i) * -1j) for i in numbers])
#     return dephase_mat



import numpy as np
from scipy.linalg import expm
import matplotlib.pyplot as plt
import scipy.stats as stats
from scipy.stats import bernoulli
import random

trans_mat = np.zeros((4, 10))
trans_mat[0, 1] = trans_mat[1, 3] = trans_mat[2, 4] = trans_mat[3, 8] = 1


def int2bin(n, count=24):
    return "".join([str((n >> y) & 1) for y in range(count-1, -1, -1)])


def loc(x):
    """
    turn a string of 0/1 into its decimal number
    :param x: string of 0/1
    :return: Decimal number
    """
    return int(x, 2)


def dagger(x):
    return x.T.conj()


def state_init(N, site):
    """
    :param N: length of state
    :param site: bit string of the site
    :type site: str
    :return: state
    """
    init_state = np.zeros(2**N)
    init_state[loc(site)] = 1
    return init_state


def W_state(N, log = False):

    coe = 1/pow(N, 0.5)
    state = np.zeros(2**N, dtype = 'complex')
    for i in range(N):
        state[2**i] = coe
    if log:
        f = open('w_psi'+str(N)+'.txt', mode='w')
        for i in range(2**N):
            f.write(str(state[i].real)+' '+str(state[i].imag)+'\n')
        f.close()
    return state


def state_save(state, path=None):
    N = int(np.log2(len(state)))
    path = path if path else str(N)+' qubits_state'+'.txt'
    f = open(path, mode='w')
    for i in range(2 ** N):
        f.write(str(state[i].real) + ' ' + str(state[i].imag) + '\n')
    f.close()
    print(path)


def amp_save(state, path=None):
    N = int(np.log2(len(state)))
    path = path if path else str(N)+' qubits_state_amp'+'.txt'
    f = open(path, mode='w')
    for i in range(2 ** N):
        f.write(str(np.abs(state[i])) + ' ' + str(0.0000) + '\n')
    f.close()
    print(path)


def sparse_check(x):
    tmp = x.flatten(order='C')
    nonzero = 0
    for i in range(len(tmp)):
        if tmp[i] != 0:
            nonzero += 1
    return nonzero, nonzero/len(tmp)


def unitary_check(x):
    threshold = 1e-10
    distance = np.linalg.norm(np.dot(dagger(x),x)-np.eye(len(x)))
    if distance<threshold:
        return True
    else:
        print('not unitary, error = {}'.format(distance))


def set_bit_val(byte, index, N, val):
    """
    更改某个字节中某一位（Bit）的值

    :param byte: 准备更改的字节原值
    :param index: 待更改位的序号，从右向左0开始，0-7为一个完整字节的8个位
    :param val: 目标位预更改的值，0或1
    :returns: 返回更改后字节的值
    """
    if val:
        return byte | (1 << (N-index))
    else:
        return byte & ~(1 << (N-index))


def site(data, N, i):
    return data >> (N-i) & 1


def fastmul(m,n, gate, state):
    N = int(np.log2(len(state)))
    index = [2*site(i,N,m)+site(i,N,n) for i in range(2**N)]
    gate = gate.T
    tmat = gate[:, index]
    v = np.arange(2**N).reshape(1,2**N).repeat(4,0)
    for i in range(4):
        p = site(i, 2, 1)
        q = site(i, 2, 2)
        v[i, :] = set_bit_val(v[i, :], m, N, p)
        v[i, :] = set_bit_val(v[i, :], n, N, q)
    v = state[v]
    tmat *= v
    res = tmat.sum(0)
    return res





def swap(U, J, t, Delta=0):
    H = np.array([[U, -np.sqrt(2)*J, 0, 0, 0, 0, 0, 0, 0,0],
              [-np.sqrt(2)*J, Delta, -np.sqrt(2)*J, 0, 0, 0, 0, 0, 0,0],
              [0, -np.sqrt(2)*J, U+2*Delta, 0, 0, 0, 0, 0, 0,0],
              [0, 0, 0, Delta, 0, -J, -J, 0, 0, 0],
              [0, 0, 0, 0, Delta, -J, -J, 0, 0, 0],
              [0, 0, 0, -J, -J, U, 0, 0, 0, 0],
              [0, 0, 0, -J, -J, 0, U+2*Delta, 0, 0, 0],
              [0, 0, 0, 0, 0, 0, 0, U, -np.sqrt(2)*J, 0],
              [0, 0, 0, 0, 0, 0, 0, -np.sqrt(2)*J, Delta, -np.sqrt(2)*J],
              [0, 0, 0, 0, 0, 0, 0, 0, -np.sqrt(2)*J, U+2*Delta]])
    Evolution = expm(H * 2*np.pi*t*-1j)
    swap = np.dot(trans_mat, Evolution)
    swap = np.dot(swap, trans_mat.T)
    swap /= 1j
    return swap


def sto(t, Delta):
    """
    only for 01/10 base
    :param t:
    :param Delta:
    :return:
    """
    phase = np.array([[0, 0, 0, 0],
                      [0, 0, 0, 0],
                      [0, 0, 2*Delta, 0],
                      [0, 0, 0, 0]])
    Evolution = expm(phase * 2*np.pi*t*-1j)
    return Evolution


def noise(x):
    mu, sigma = x, x/25.76
    np.random.seed()
    return np.random.normal(mu, sigma, 1)[0]


def swap_noise(U, J, t, Delta = 0):
    #     U, J, t, Delta = noise(U), noise(J), noise(t), Delta = noise(Delta)
    return swap(noise(U), noise(J), noise(t), noise(Delta))


def sto_noise(t, Delta):
    return sto(noise(t), noise(Delta))


def NumberOf1(n):
    count = 0
    while n&0xffffffff != 0:
        count += 1
        n = n & (n-1)
    return count


def phase_shift(n):
    phase = 0
    for _ in range(n):
        phase += noise(2*np.pi)
    return phase


def dephase(n):
    numbers = []
    for i in range(2 ** n):
        numbers.append(NumberOf1(i))
    numbers = np.array(numbers)
    dephase_mat = np.diag([np.exp(phase_shift(i)*-1j) for i in numbers])
    return dephase_mat


def density_mat(x):
    n = len(x)
    sam_shape = len(x.shape)
    if sam_shape == 1:
        dim = len(x)
        state = x.reshape(1, dim)
        state /= np.linalg.norm(state)
        den_mat = np.dot(dagger(state), state)
    else:
        dim = len(x[0])
        den_mat = np.zeros((dim, dim))
        for i in range(n):
            state = x[i].reshape(1, dim)
            state /= np.linalg.norm(state)
            if not i:
                den_mat = np.dot(dagger(state), state)
            else:
                den_mat += np.dot(dagger(state), state)
        den_mat /= n
    return den_mat


def fidelity_vec(x, y):
    return (np.dot(x.conj(), y)*np.dot(y.conj(), x)/np.linalg.norm(x)**2/np.linalg.norm(y)**2).real


def fidelity_essemble(x,y):
    n = len(y)
    fidelity = 0
    for i in range(n):
        fidelity += fidelity_vec(x, y[i])
    return fidelity/n


def purity(x):
    return (np.trace(np.dot(x, x))).real


def distribution(x):
    x /= np.linalg.norm(x, 2)
    prob = np.zeros(x.size)
    for i in range(len(x)):
        prob[i] = np.abs(x[i])**2
    return prob


def dis2state(x):
    state = np.array([np.sqrt(x[i]) for i in range(len(x))])
    return state


def sample(x, n=1):
    N = int(np.log2(len(x)))
    res = stats.rv_discrete(values=(range(len(distribution(x))), distribution(x))).rvs(size=n)
    if n == 1:
        return res[0]
    else:
        dis = sample_distribution(res, N)
        kl = KL(distribution(x), dis)
        print(kl)
        return res, dis


def sample_distribution(sample, N):
    n = len(sample)
    sample_dis = np.zeros(2**N)
    for i in sample:
        sample_dis[i] += 1
    sample_dis /= n
    return sample_dis


def essemble_distribution(x):
    n = len(x)
    dis = np.array([x[i][i] for i in range(n)])
    return dis.real


def KL(p, q):
    divergence = 0
    for i in range(len(p)):
        if p[i] and q[i]:
            divergence += p[i]*np.log(p[i]/q[i])
    return divergence


def KL_new(P,Q):
    N = len(P)
    epsilon = 0.01/N
    P = P + epsilon
    Q = Q + epsilon
    divergence = np.sum(P*np.log(P/Q))
    return divergence


def sample_plot(dis, N, M, KL=None):
    x = [int2bin(i,N) for i in range(len(dis))]
    plt.bar(x, dis)
    plt.ylim(0,1)
    for x, y in enumerate(dis):
        plt.text(x, y+0.02, '%s' %y, ha='center')
    if not KL:
        plt.title('{} qubits with {} measurements'.format(N, M))
    else:
        plt.title('{} qubits with {} measurements\n KL = {}'.format(N, M, KL))
    plt.ylabel('Probility')
    plt.show()


def trans_base(bases, x):
    Z2Z = np.eye(2)
    Z2X = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
    Z2Y = 1 / np.sqrt(2) * np.array([[1, -1j], [1, 1j]])
    decode = {'Z': Z2Z, 'X': Z2X, 'Y': Z2Y, 'z': Z2Z, 'x': Z2X,'y': Z2Y}
    tmp_mat = decode[bases[0]]
    for i in range(1,len(bases)):
        tmp_mat = np.kron(tmp_mat, decode[bases[i]])
    return np.dot(tmp_mat, x)


def sample_bases(bases, state, M):
    N = int(np.log2(len(state)))
    f1 = open(str(N)+' qubits_measurement.txt', mode='w')
    f2 = open(str(N)+' qubits_measurement_bases.txt', mode='w')
    f3 = open(str(N)+' qubits_measurement_bases_set.txt', mode='w')
    for i in bases:
        measure = sample(trans_base(i, state), M)
        for j in measure:
            tmp = int2bin(j, N)
            for k in tmp:
                f1.write(k+' ')
            f1.write('\n')
            for k in i:
                f2.write(k+' ')
            f2.write('\n')
        for j in i:
            f3.write(j+' ')
        f3.write('\n')
    f1.close()
    f2.close()
    f3.close()


def Z_sample(state, M, error=0):
    N = int(np.log2(len(state)))
    f1 = open(str(N)+' qubits_measurement_z.txt', mode='w')
    measure = sample(state, M)
    dis = sample_distribution(measure, N)
    kl = KL(dis, distribution(state))
    print(kl)
    if error:
        measure = sample_error(measure, N, error)
        dis = sample_distribution(measure, N)
        kl = KL(dis, distribution(state))
        print(kl)
    for j in measure:
        tmp = int2bin(j, N)
        for k in tmp:
            f1.write(k+' ')
        f1.write('\n')
    f1.close()


def sample_error(samples, n, error):
    size = len(samples)
    flip = bernoulli.rvs(n * error, size=size)
    ker = [2 ** i for i in range(n)]
    count = 0
    for i in range(size):
        if flip[i]:
            count += 1
            flip[i] = random.choice(ker)
    print(count, count / size)
    sample_new = np.array([samples[i] ^ flip[i] for i in range(size)])
    return sample_new


def sample_save(samples, N, path=None):
    path = path if path else str(N)+' qubits_measurement_z.txt'
    f1 = open(path, mode='w')
    for j in samples:
        tmp = int2bin(j, N)
        for k in tmp:
            f1.write(k+' ')
        f1.write('\n')
    f1.close()
    print(path)


import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="qusource",
  version="0.0.1",
  author="Xiangrui Meng",
  author_email="mxr@mail.ustc.edu.cn",
  description="Quantum circuit simulator",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/Meng-Xiang-Rui/Qusource",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
  python_requires='>=3.6',
  install_requires=[
      'numpy',
  ],
)


import numpy as np


def loc(x):
    """
    turn a string of 0/1 into its decimal number
    :param x: string of 0/1
    :return: Decimal number
    """
    return int(x, 2)


def length(state):
    """
    the length of the atoms chain of state.
    :param state: the quantum state
    :return: the length
    """
    return int(np.log2(len(state)))


def site_bit(number, N, site):
    """
    0/1 bit of the i-th atom in the base of N atoms chain from left-side
    :param number: state base
    :param N: length of atoms chain
    :param site: index of the atom in the atoms chain
    :return: 0/1
    """
    return number >> (N-site) & 1


def index(number, N, sites):
    """
    the index of number in the subspace generated by qubits in sites
    :param number: number
    :param N: length of atoms chain
    :param sites: sites of qubits
    :type sites: array-like
    :rtype: int
    """
    n = len(sites)
    res = 0
    for i in range(n):
        res += site_bit(number, N, sites[n - 1 - i]) * 2 ** i
    return res


def int2bin(n, count=24):
    """
    trans a number to binary string
    :param n: number
    :type n: int
    :param count: total length
    :type count: int
    :rtype: binary string
    """
    return "".join([str((n >> y) & 1) for y in range(count-1, -1, -1)])


def set_bit(number, site, N, bit):
    """
    Change the bit value of number at a single site.

    :param number: the number to be changed
    :param site: the index of the site starting from left
    :param bit: the bit value to be changed
    :returns: the changed number
    """
    if bit:
        return number | (1 << (N-site))
    else:
        return number & ~(1 << (N-site))
